module nr
!	use nrtype
	IMPLICIT NONE

CONTAINS

SUBROUTINE amoeba(p,y,ftol,func,iter)
!use nrtype; 
use nrutil, ONLY : assert_eq, imaxloc, iminloc, nerror, swap
IMPLICIT NONE
Integer , intent(out) :: iter
Double Precision, intent(in)  :: ftol
Double Precision, dimension(:), intent(inout) :: y
Double Precision, dimension(:,:), intent(inout) :: p

Interface
	Function func(x)
		!use nrtype
		IMPLICIT NONE
		Double Precision, dimension(:), intent(in) :: x
		Double Precision :: func
	end Function func
end Interface

Integer, parameter :: ITMAX = 5000
!	Mininization of the function, func, in N dimensions by the downhill 
!	simplex method of Nelder and Mead.  The (N+1) x N maxtrix p is input. Its
!	N+1 rows are N-dimensional vectors that are the vertices of the starting
!	simplex.  Also input is the vector y of the length N+1, whose components
!	must be preinitialized to the values of func evaluated to at the
!	N+1 vertices (rows) of p; and ftol the fractional convergence tolerance
!	to be achieved in the FUNCTION VALUE.  On output, p and y will have
!	been reset to N+1 new points all within ftol of a minimum function value,
!	and iter gives the number of function evaluations taken.
!	Parameter: The maximum allowed number of function evaluations.

Integer :: ihi, ndim								!Global variables
Double Precision, dimension(size(p,2))  :: psum
CALL amoeba_private

Contains

Subroutine amoeba_private
IMPLICIT NONE
Integer  :: i, ilo, inhi
Double Precision :: rtol, ysave, ytry, ytmp

	ndim=assert_eq(size(p,2),size(p,1)-1,size(y)-1,'amoeba')
	iter = 0
	psum(:) = sum(p(:,:),dim=1)
	do
		ilo    = iminloc(y(:))  !determine which point is highest (worst)
		ihi    = imaxloc(y(:))  !next highest and lowest (best)
		ytmp   = y(ihi)
		y(ihi) = y(ilo)
		inhi   = imaxloc(y(:))
		y(ihi) = ytmp
		print *,"Iteration:   ",iter
      print '(A,E24.12)',"Minimum [y(ilo)] ",y(ilo)
      print '(A,E24.12)',"Maximum [y(ihi)] ",y(ihi)
      print '(A)',"   Parameters at low and high points."
      print '(3(3F24.12,/))',p(ilo,:)
		print *," "
      print '(3(3F24.12,/))',p(ihi,:)
      print *," "
		rtol   = 2.0*abs(y(ihi)-y(ilo))/(abs(y(ihi))+abs(y(ilo)))
		!compute fraction range from highest to lowest and return if satisfactory
		if (rtol < ftol) then	!if returning put best point and value in slot 1
			call swap(y(1),y(ilo))
			call swap(p(1,:),p(ilo,:))
			RETURN
		end if
		if (iter >= ITMAX) call nerror('ITMAX exceeded in amoeba')
		!begin new iteration.  First extrapolate by a factor -1 through the
		!face of the simplex across from the highest point, i.e., reflect
		!the simplex from the high point.
		ytry=amotry(-1.0d0)
		if (ytry <= y(ilo)) then	!gives result better than the best point
			ytry = amotry(2.0d0)	!so try an additional extrapolation by
			iter = iter + 1			!a factor of 2
		else if (ytry >= y(inhi)) then !the reflected point is worse than the
			ysave=y(ihi)					 !2nd highest, so look for an intermediate
			ytry = amotry(0.5d0)		 !lower pt, i.e., do a one-dimensional 
			iter = iter+1					 !contraction
			
			if (ytry >= ysave) then	!cant seem to get rid of that high point. 
											!Better contract around the lowest (best)
											!point.
				p(:,:) = 0.5*(p(:,:)+spread(p(ilo,:),1,size(p,1)))
				do i=1,ndim+1
					if (i /= ilo) y(i) = func(p(i,:))
				end do
				iter = iter+ndim	!keep track of function evaluations
				psum(:) = sum(p(:,:),dim=1)
			end if
		end if
	end do     !go back for the test of doneness and the next iteration
end SUBROUTINE amoeba_private

Function amotry(fac)
IMPLICIT NONE
Double Precision, intent(in) :: fac
Double Precision :: amotry
!Extrapolates by a factor, fac, through the face of the simplex across from
!the highest point, tries it, and replaces the high point if the new point
!is better.
	Double Precision :: fac1, fac2, ytry
	Double Precision , dimension(size(p,2)) :: ptry

	fac1 = (1.0-fac)/ndim
	fac2 = fac1-fac
	ptry(:) = psum(:)*fac1-p(ihi,:)*fac2
	ytry=func(ptry)		      !evaluate function at trial point
	if (ytry < y(ihi)) then    !If it's better than the highest, then replace
		y(ihi) = ytry				!the highest.
		psum(:) = psum(:) - p(ihi,:) + ptry(:)
		p(ihi,:)= ptry(:)
	end if
	amotry = ytry
end Function amotry
end SUBROUTINE amoeba
		
FUNCTION brent(ax,bx,cx,func,tol,xmin)
!use nrtype; 
use nrutil, ONLY : nerror
IMPLICIT NONE
Double Precision , intent(in) :: ax,bx,cx,tol
Double Precision , intent(out) :: xmin
Double Precision ::  brent

interface
	FUNCTION func(x)
		!use nrtype
		IMPLICIT NONE
		Double Precision , intent(in) :: x
		Double Precision :: func
	end FUNCTION func
end interface

	Integer , parameter :: ITMAX=100
	Double Precision , parameter :: CGOLD=0.3819660, ZEPS = 1.0e-3*EPSILON(ax)

!	Given a function, func, and given a bracketing triplet of abscissas ax, 
!	bx, cx (such that bx is between ax and cx, and func(bx) is less than
!	both func(ax) and func(cx)), this routine isolates the minimum to a
!	fractional precision of about tol using Brent's method.
!
!	The abscissa of the minimum is returned in xmin and the minimum function
!	value is returned as brent, the returned function value.
!
!	Parameters: Maximum allowed number of iterations; golden ratio; and a 
!	small number that protects against trying to achieve accuracy for a 
!	minimum that happens to be zero.

	Integer :: iter
	Double Precision :: a,b,d,e,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm
	
!a,b must be in ascending order, thou input abscissas need not be
	a=min(ax,cx)			
	b=max(ax,cx)
	v=bx
	w=v
	x=v
	e=0.0				!This will be distance on step before last
	fx=func(x)
	fv=fx
	fw=fx
!	print *,"Brent iterations:"
!	print *,"Iter                       x          function value"
	do iter=1,ITMAX
!		print '(I5,E24.16,E24.16)',iter,x,fx
		xm=0.5 * (a+b)
		tol1 = tol*abs(x) + ZEPS
		tol2 = 2.0*tol1
		if (abs(x-xm) <= (tol2-0.5*(b-a))) then 	!test for done here
			xmin = x				! arrive here ready to exit with best values
			brent = fx
			RETURN
		end if
		if ( abs(e) > tol1 ) then	!construct a parabolic fit
			r=(x-w)*(fx-fv)
			q=(x-v)*(fx-fw)
			p=(x-v)*q-(x-w)*r
			q=2.0*(q-r)
			if (q > 0.0) p=-p
			q=abs(q)
			etemp=e
			e=d
!	The next condition determines the acceptability of the parabolic fit.
			if (abs(p) >= abs(0.5*q*etemp) .or. &
				p <= q*(a-x) .or. p >= q*(b-x) ) then

!	Here it is not o.k., so we take the golden section step into the 
!	larger of the two segments.

				e=merge(a-x,b-x,x >= xm)
				d=CGOLD*e
			else			!take the parabolic step
				d=p/q
				u=x+d
				if (u-a < tol2 .or. b-u < tol2) d=sign(tol1,xm-x)
			end if
		else 
			e=merge(a-x,b-x, x >= xm) !take the golden step in larger of two segs
			d=CGOLD*x
		end if

		u=merge(x+d,x+sign(tol1,d), abs(d) >= tol1 )

!	arrive here with d computed either from parabolic fit or else from 
!	golden section

		fu=func(u)		!the one function evaluation per iteration 

		if (fu <= fx) then  !housekeeping
			if (u >= x) then
				a=x
			else
				b=x
			end if
			call shft(v,w,x,u)
			call shft(fv,fw,fx,fu)
		else
			if (u < x) then
				a = u
			else
				b = u
			end if
			if (fu <= fw .or. w == x) then
				v=w
				fv=fw
				w=u
				fw=fu
			else if (fu <= fv .or. v == x .or.  v == w) then
				v=u
				fv=fu
			end if
		end if
	end do
	call nerror('Brent: Exceed maximum iterations.')


	CONTAINS

	SUBROUTINE shft(a,b,c,d)
	Double Precision , intent(out) :: a
	Double Precision , intent(inout) :: b,c
	Double Precision , intent(in) :: d
		a=b
		b=c
		c=d
	end SUBROUTINE shft
end FUNCTION brent
	
SUBROUTINE linmin(p,xi,fret)
	!Use nrtype; 
	use nrutil, ONLY : assert_eq
!	Use nr, Only : mnbrak, brent
	Use f1dim_mod
	IMPLICIT NONE

	Double Precision, intent(out) :: fret
	Double Precision, intent(inout), dimension(:), target :: p, xi
	Double Precision, parameter :: TOL=1.0e-4
!	Given an N-dimensional point p and an N-dimensional direction xi, both
!	vectors of length N, moves and resets p to where the fixed-name function
!	func takes on a minimum along the direction xi from p, and replaces xi
!	by the actual vector displacement that p was moved.  Also returns as
!	fret the value of funct at the returned function location p.  This is
!	actually all accomplished by calling the routines mnbrak and brent.
!	Parameter :: Tolerance passed to brent
!
	Double Precision :: ax,bx,fa,fb,fx,xmin,xx
	
	ncom = assert_eq(size(p),size(xi),'linmin')
	
!	Communicate the global variables to f1dim
	pcom => p
	xicom => xi

!	initial guess for brackets (may want to change)
	ax = 0.0
	xx = 1.0
	print *,"Step 1: mnbrak-Bracket minimum"
	Call mnbrak(ax,xx,bx,fa,fx,fb,f1dim)
	fret = brent(ax,xx,bx,f1dim,TOL,xmin)
!	construct the vector results to return
	xi = xmin*xi
	p = p+xi
	print *,"linmin:: new direction and new p"
	print '(A,2E24.12)',"New p: ",p
end SUBROUTINE linmin

SUBROUTINE mnbrak(ax,bx,cx,fa,fb,fc,func)
!use nrtype; 
use nrutil, ONLY : swap 
IMPLICIT NONE
Double Precision, intent(inout) :: ax, bx
Double Precision, intent(out)	:: cx, fa,fb,fc

Interface
	FUNCTION func(x)
	!Use nrtype 
	IMPLICIT NONE
	Double Precision, intent(in) 	:: x
	Double Precision :: func
	end FUNCTION func
end interface

Double Precision , parameter :: GOLD=1.618034, GLIMIT=100.0, TINY=1.0e-20

Double Precision :: fu,q,r,u,ulim

!Given a function FUNC and given a distinct initial points ax and bx, this 
!routine searchs in the downhill direction (defined by the function as 
!evaluated at the initial points) and the function values at the three points
!fa,fb,fc.

!Parameter GOLD is the default ratio by which successive intervals are
!magnified.  GLIMIT is the maximum magnification allowed for a parabolic-fit
!step.

	fa=func(ax)
	fb=func(bx)
!	Switch a, b to go downhill in direction from a to b. 
	if (fb > fa) then
		call swap(ax,bx)
		call swap(fa,fb)
	end if
	cx = bx + GOLD*(bx-ax)
	fc = func(cx)
	do								! return to here until we bracket
		if (fb < fc) return 

!	Compute u by parabolic extrapolation from a,b,c.  TINY used to prevent
!	any possible division by zero.

		r=(bx-ax)*(fb-fc)
		q=(bx-cx)*(fb-fa)
		u=bx-( (bx-cx)*q -(bx-ax)*r ) / (2.0*sign(max(abs(q-r),TINY),q-r))
		ulim=bx+GLIMIT*(cx-bx) 
!won't go farther than this;  test possibilities
		if ( (bx-u)*(u-cx) > 0.0) then !parabolic u betw b,c, try it.
			fu=func(u)
			if (fu < fc) then				 !got a min betw b,c
				ax=bx
				fa=fb
				bx=u
				fb=fu
				return
			else if (fu > fb) then      !got a min betw a, u
				cx=u
				fc=fu
				return
			end if
			u=cx+GOLD*(cx-bx)				!para fit of no use, use default mag 
			fu=func(u)
			else if ( (cx-u)*(u-ulim) > 0.0) then !para fit betw c, and limit
				fu=func(u)
				if (fu < fc) then
					bx=cx
					cx=u
					u=cx + GOLD*(cx-bx)
					call shft(fb,fc,fu,func(u))
				end if
			else if ( (u-ulim)*(ulim-cx) >= 0.0) then !limit para u to max limit
				u=ulim
				fu=func(u)
			else										!rej para u, use default mag
				u=cx+GOLD*(cx-bx)
				fu=func(u)
			end if
			call shft(ax,bx,cx,u)
			call shft(fa,fb,fc,fu)
		end do

CONTAINS
	SUBROUTINE shft(a,b,c,d)
	IMPLICIT NONE
		Double Precision, intent(out) :: a
		Double Precision, intent(inout) :: b,c
		Double Precision, intent(in)  :: d
		a = b
		b = c
		c = d
	end SUBROUTINE shft

end Subroutine mnbrak 

Subroutine powell(p,xi,ftol,iter,fret)
	!use nrtype; 
	use nrutil, only : assert_eq, nerror
!	use nr, ONLY : linmin
	IMPLICIT NONE
	Double Precision, dimension(:), intent(inout) :: p
	Double Precision, dimension(:,:) , intent(inout) :: xi
	Integer , intent(out) :: iter
	Double Precision , intent(in) :: ftol
	Double Precision , intent(out) :: fret

	Interface
		Function func(p)
			!use nrtype; 
			IMPLICIT NONE
			Double Precision , intent(in), dimension (:) :: p
			Double Precision :: func
		end Function func
	end interface

	Integer , parameter :: ITMAX = 200
!
!	Minimization of a function func of N variables (func is not an argument,
!	it is a fixed function name).  Input consits of an initial starting point
!	p, a vector of length N, an initial NxN matrix, xi, whose columns contain
!	the initial set of directions (usually the NxN unit vectors), and ftol
!	the fractional tolerance value such that failure to decrease by more than
!	this amount on one iteration signals doneness.  On output, p, is set to
!	the best point found, xi, is the then-current direction set, fret is the
!	returned function value at p, and iter is the number of iterations taken.
!	Function linmin used.
!
!	Parameter:: maximum allowed iterations.
!	
!	Routine from NR in Fortran 90, pp 1210-1212.
!
	Integer :: i, ibig, n
	Double Precision :: del, fp, fptt, t
	Double Precision , dimension(size(p)) ::  pt, ptt, xit

	n = assert_eq(size(p),size(xi,1),size(xi,2),'powell')
	fret = func(p)
	pt(:) = p(:)	!save initial point
	iter = 0
	do
		iter = iter+1
		fp = fret
		ibig = 0
		del = 0.0
		do i=1,n						!will be the biggest function decrease
			xit(:) = xi(:,i)		!Loop over all directions in the set.
			fptt = fret				!Copy the direction
			print *,"Powell: minimize for parameter i=",i
			CALL linmin(p,xit,fret) !minimize along it
			if (abs(fptt-fret) > del) then	!record it if is the largest decrease
				del = abs(fptt-fret)				!so far
				ibig = i
			end if
		end do
!	check termination criterion
		if (2.0*abs(fp-fret) <= ftol*(abs(fp)+abs(fret))) RETURN

		if (iter == ITMAX) CALL &
			nerror('powell exceeding maximum iterations')
!
!	Construct the extrapolated point and the average direction moved. Save
!	the old starting point.

		ptt(:) = 2.0*p(:) - pt(:)  
		xit(:) = p(:) - pt(:)
		pt(:)  = p(:)

!	Obtain function value at extrapolated point.
		fptt = func(ptt)

!	one reason not to use the new point
		if (fptt >= fp) cycle

!	another reason not to use the new direction
		t=2.0*(fp-2.0*fret+fptt)*(fp-fret-del)**2-del*(fp-fptt)**2
		if (t >= 0.0) cycle

!	Move to minimum of the new direction, and save the new direction
		CALL linmin(p,xit,fret)
		xi(:,ibig) = xi(:,n)
		xi(:,n)=xit(:)

!	Back for another iteration
	end do			

end Subroutine powell
end module nr
